%%This is a part of multi-layered random forest (MLRF) framework

%Copyright (C) 2014 Alina Kuznetsova

%This program is free software: you can redistribute it and/or modify
%it under the terms of the GNU Lesser General Public License as published by
%the Free Software Foundation, either version 2.1 of the License, or
%(at your option) any later version.

%This program is distributed in the hope that it will be useful,
%but WITHOUT ANY WARRANTY; without even the implied warranty of
%MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%GNU Lesser General Public License for more details.

%You should have received a copy of the GNU General Public License
%along with this program.  If not, see <http://www.gnu.org/licenses/>.

%Author: Alina Kuznetsova
    
%% train a multi-layered random forest
%%Input:
%	X - train feature vectors Nxdim
%	Y - train responce 1xN or Nx1
%	clusters - data clustering, 1xNcl or Nclx1, Ncl <= N (Ncl<N -> the rest of
%the data will be classified using the 1st layer of the forest before
%training the second layer)
%	options - various options
%%Output:
%   classifier - classifier description
%   clusters - clusters (if Ncl== N), clusters(in)==clusters(out), else, clusters are
%determined for the whole training set
%   del - removed training samples (empty, if none was removed)


function [classifier,clusters,del] = train(X,Y,clusters,options)

if(~isfield(options,'pathIsSet'))
    addpath(genpath('../MLRF'));
end

N = size(X,1);
dim = size(X,2);

Ncl = length(clusters);

if (size(Y,1) < size(Y,2))
    Y = Y'; %make a column! 
end

if (size(clusters,1) < size(clusters,2))
    clusters = clusters';
end

if (N~= length(Y))
    disp(size(X))
    disp(size(Y))
    disp(size(clusters));
    error('Y has the wrong dimensions');
end

if (N < length(clusters))
    %first length(clusters) components are for clustering
    error('clusters has the wrong dimenstions');
end

%first train clustering classifier
ucl = unique(clusters);
clusterCount = length(ucl);
classCount = length(unique(Y));

if(~isfield(options,'verbose') || options.verbose)
    options.verbose = 1;
    disp('Training...');
    disp(['Feature vector dimensions: ' num2str(dim)]);
    disp(['Number of samples: ' num2str(N)]);
    disp(['Number of clusters: ' num2str(clusterCount)]);
    disp(['Number of classes: ' num2str(classCount)]);
    disp('Options: ');
    disp(options);
end

%convert Y to numeric representation if needed and add the encoding
if (~isnumeric(Y))
    [Y,code] = char2num(Y);
    classifier.code = code;
end

if (isfield(options.clusterOptions,'aggLevel'))
    Xagg = featureAggregation(X,options.clusterOptions.aggLevel);
else
    Xagg = X;
end

%normalize features to fit to N(0,1)
if (options.normalize)

    [X,Xmean,Xstd,exclude] = normalize(X,options);

    classifier.exclude = exclude;
    classifier.Xstd = Xstd;
    classifier.Xmean = Xmean;
    
    if (isfield(options.clusterOptions,'aggLevel'))
        [Xagg,Xmean,Xstd,exclude] = normalize(Xagg,options);
        
        classifier.XaggExclude = exclude;
        classifier.XaggStd = Xstd;
        classifier.XaggMean = Xmean;
    end
    
end

if (clusterCount > 1)
   %train a set of classifiers 
    warning(['Cluster number: ' num2str(clusterCount)]);
    %train a forest to determine a cluster
    %RF forest options should be different for different levels
    %or use KNN?
    
    [m,err,CTrain] = trainClassifier(Xagg(1:Ncl,:),clusters,options.RFLevel0options);
    
    classifier.type = 'hierarchy';
    classifier.clusters = ucl;
    
    level0.type = 'RF';
    level0.object = m;
    level0.err = err;
    level0.confMat = CTrain;
    
    classifier.level0 = level0;
    
    if (Ncl<N)
        %classify the rest using the RF to increase the amount of training
        %data
        Ycl = forestTest(level0.object, Xagg(Ncl+1:end,:));
    
        clusters = [clusters; Ycl];
        
        %class set should stay the same  - yet
        
        del = removeSmallClasses(Y,clusters,confusionMatrix(clusters,Y),options.clusterOptions);
        
        %small test for correctness

        classifier.notUsed = sum(del)/length(clusters);
        
        warning(['through away ' num2str(sum(del)) ' samples from the training set of size ' num2str(length(clusters))]);
        
        clusters(del) = [];
        
        if (length(unique(clusters))~=clusterCount)
            error('number of clusters is different from the initial number of clusters');
        end        
        
        X(del,:) = [];
        Y(del) = [];
    else
        del = [];
    end  
    
    if (~options.clusteringOnly)
    
        classifier.level1 = [];
    
        Yt = zeros(size(Y));
    
        for i= 1:clusterCount
            %select samples for training
            Xsub = X(clusters == ucl(i),:);
            Ysub = Y(clusters == ucl(i));
            
            if (options.normalizeClusters)
                 %renormalize - make the values range even smaller!
                 [Xsub,Xmean,Xstd,exclude] = normalize(Xsub,options);
                 cl1.Xmean = Xmean;
                 cl1.Xstd = Xstd;
                 cl1.exclude = exclude;
            end            
           
            if (length(unique(Ysub))>1)
        
                [m,err,CTrain,Ytsub] = trainClassifier(Xsub,Ysub,options.RFLevel1options);
        
                Yt(clusters == ucl(i)) = Ytsub;
        
                cl1.type = 'RF';
                cl1.object = m;
                cl1.err = err;
                cl1.confMat = CTrain;
                
            else
                cl1.type = 'const';
                cl1.object = Ysub(1);
                cl1.err = 0;
                cl1.confMat = confusionMatrix(Ysub,Ysub);
                
                Yt(clusters == ucl(i)) = Ysub(1);
            end
            
            classifier.level1 = [classifier.level1 cl1];
        end
    
        classifier.fullTrainConfMat = confusionMatrix(Y,Yt);
    end
else
    [m,err,CTrain] = trainClassifier(X,Y,options.RFLevel0options);
    classifier.type = 'RF';
    classifier.object = m;
    classifier.err = err;
    classifier.fullTrainConfMat = CTrain;
    del = [];
end

end

function [m,err,CTrain,Yt] = trainClassifier(X,Y,options)

    disp(['Train forest; features vector size: ' num2str(size(X))]);

    m = forestTrain(X, Y, options);
    
    Yt = forestTest(m, X);
    
    if (size(Yt,1) < size(Yt,2))
        Yt = Yt';
    end
    
    err = sum(Yt~=Y);
    
    CTrain = confusionMatrix(Y,Yt);

end
